package ch.hslu.appe.inventory.logic;

import java.util.Date;
import java.util.List;

import ch.hslu.appe.inventory.logic.exception.ReservationNotExistsException;
import ch.hslu.appe.inventory.logic.exception.ReservationNotPossible;
import ch.hslu.appe.inventory.model.Product;

/**
 * Inventory / Stock System Interface.
 * @author Felix Rohrer <felix.rohrer@stud.hslu.ch>
 */
public interface InventorySystem {
    /**
     * Submit a Product Order, if required a stock order will be executed.
     * @param reservationId the reservation id
     * @return Returns the estimated product delivery date
     */
    Date commitProductReservation(int reservationId);

    /**
     * Deletes a product reservation.
     * @param reservationId The reservation id
     * @return true on success, otherwise false
     * @throws ReservationNotExistsException no Reservation found
     */
    boolean deleteProductReservation(int reservationId) throws ReservationNotExistsException;

    /**
     * Convert ProductNumber from int to String (fixed length: 6 char).
     * @param productNumber productNumber as int
     * @return productNumber as String
     */
    String generateStockProductNumber(final int productNumber);

    /**
     * Returns a list with all active Products.
     * @return a list with all active Products.
     */
    List<Product> getActiveProducts();

    /**
     * Returns a Product based on ProductNumber.
     * @param productNumber requested ProductNumber
     * @return Product
     */
    Product getProduct(int productNumber);

    /**
     * Returns a list with all Products.
     * @return a list with all Products.
     */
    List<Product> getProducts();

    /**
     * Returns if a Product is active (true) or not (false).
     * @param productNumber requested ProductNumber
     * @return true if active otherwise false
     */
    boolean productActive(int productNumber);

    /**
     * Returns if a Product exists (true) or not (false).
     * @param productNumber requested ProductNumber
     * @return true if exists otherwise false
     */
    boolean productExists(int productNumber);

    /**
     * Initiates a product reservation.
     * @param productNumber The product number
     * @param amount The amount
     * @return Returns the reservation on success, otherwise -1
     * @throws ReservationNotPossible ReservationNotPossible
     */
    int reservateProduct(int productNumber, int amount) throws ReservationNotPossible;

    /**
     * Update the product.
     * @param product The product
     * @return Returns true on success, otherwise false
     */
    boolean updateProduct(Product product);

    /**
     * Updates an existing product reservation.
     * @param reservationId The reservation id
     * @param amount The new amount
     * @return true on success, otherwise false
     * @throws ReservationNotExistsException no Reservation found
     */
    boolean updateProductReservation(int reservationId, int amount) throws ReservationNotExistsException;

    /**
     * Updates the Stock, StockCount.
     * @param productNumber the product-number
     * @param newStock new Stock info
     * @return true if success otherwise false
     */
    boolean updateStock(int productNumber, int newStock);
}
