package ch.hslu.appe.inventory.logic;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import ch.hslu.appe.inventory.logic.exception.ReservationNotExistsException;
import ch.hslu.appe.inventory.logic.exception.ReservationNotPossible;
import ch.hslu.appe.inventory.logic.model.ProductReservation;
import ch.hslu.appe.inventory.model.Product;
import ch.hslu.appe.inventory.model.StockOrder;
import ch.hslu.appe.inventory.persistency.InventoryPersistency;
import ch.hslu.appe.inventory.persistency.InventoryPersistencyImpl;
import ch.hslu.appe.order.model.OrderState;
import ch.hslu.appe.stock.Stock;
import ch.hslu.appe.stock.StockException;
import ch.hslu.appe.stock.StockFactory;

/**
 * Inventory / Stock System Implementation.
 * @author Felix Rohrer <felix.rohrer@stud.hslu.ch>
 */
public final class InventorySystemImpl implements InventorySystem {

    private static final InventoryPersistency INVENTORY_SYSTEM = new InventoryPersistencyImpl();
    private static final Logger LOGGER = LogManager.getLogger("InventorySystemLogic");

    private static final ArrayList<ProductReservation> RESERVATIONS = new ArrayList<ProductReservation>();

    @Override
    public Date commitProductReservation(final int reservationId) {
        Date deliverDate = null;

        final ProductReservation reservation = getProductReservation(reservationId);
        if (reservation == null) {
            return null;
        }

        final Product product = INVENTORY_SYSTEM.getProduct(reservation.getProductNumber());
        final int amountLocalAvailable = product.getAmount();
        final int amountLocalMinAvailable = product.getAmountmin();
        final int amountOrdered = reservation.getAmount();
        int amountStockOrdered = 0;

        // do we need to place a stock order?
        // Bsp: Ava: 50, Min: 10, Order 70 --> 50 - 70 = -20 < 10 = true --> do StockOrder(30)
        // Ava: 50, Min: 10, Order 41 --> 50 - 41 = 9 < 10 = true --> do StockOrder(19)
        // Ava: 50, Min: 10, Order 10 --> 50 - 10 = 40 < 10 = false --> we can serve them
        // Ava: 21, Min: 14, Order 7 --> 21 - 7 = 14 < 14 = false
        final int amountStockOrder = amountLocalAvailable - amountOrdered;
        if (amountStockOrder < amountLocalMinAvailable) {
            amountStockOrdered = Math.abs(amountStockOrder) + (2 * amountLocalMinAvailable);

            deliverDate = doStockOrder(reservation.getProductNumber(), amountStockOrdered);
            if (deliverDate == null) {
                return null; // StockOrder failed!
            }
        } else {
            LOGGER.trace("No StockOrder required, we can serve them by ourself.");

            deliverDate = new Date();
        }

        int newStockAmount = 0;
        if (amountLocalAvailable - amountOrdered > 0) {
            newStockAmount = amountLocalAvailable - amountOrdered;
        }

        final boolean updateStockSuccess = INVENTORY_SYSTEM.updateStock(product.getNumber(), newStockAmount);
        if (!updateStockSuccess) {
            LOGGER.error("Local Stock could not be updated!");
        }

        // remove reservation
        try {
            deleteProductReservation(reservationId);
        } catch (final ReservationNotExistsException e) {
            // do nothing - we have already checked if this reservation exists!
            LOGGER.error("Reservation not found! Should never been the case!", e);
        }

        LOGGER.trace("commitProductReservation DeliverDate: " + deliverDate);
        return deliverDate;
    }

    @Override
    public boolean deleteProductReservation(final int reservationId) throws ReservationNotExistsException {

        synchronized (RESERVATIONS) {
            final ProductReservation reservation = getProductReservation(reservationId);

            if (reservation == null) {
                throw new ReservationNotExistsException(reservationId);
            }

            LOGGER.trace("ProductReservation '" + reservationId + "' successfully deleted");

            return RESERVATIONS.remove(reservation);
        }
    }

    /**
     * Execute a StockOrder.
     * @param productNumber requested ProductNumber
     * @param amount requested amount of this product
     * @return DeliverDate
     */
    private Date doStockOrder(final int productNumber, final int amount) {
        // Cast our internal ProductNumber(int) to a String for the Stock
        final String stockProductNumber = generateStockProductNumber(productNumber);

        final Stock stock = StockFactory.getStock();
        int amountStockDelivers = 0;

        try {
            amountStockDelivers = stock.orderItem(stockProductNumber, amount);
        } catch (final StockException e) {
            amountStockDelivers = 0;
            LOGGER.error("Stock orderItem Error!", e);
        }

        if (amountStockDelivers < 0) {
            LOGGER.trace("Stock order failed - deliverDate unknown!");
            return null;
        }

        // get deliverDate
        Date deliverDate;
        try {
            deliverDate = stock.getItemDeliveryDate(stockProductNumber);
        } catch (final StockException e) {
            LOGGER.error("Stock check DeliverDate Error!", e);
            return null;
        }

        // save StockOrder
        final Calendar cal = Calendar.getInstance();
        cal.setTime(new Date());
        final Date orderDate = cal.getTime();

        final StockOrder stockOrder = new StockOrder();
        stockOrder.setAmount(amountStockDelivers);
        stockOrder.setProductId(getProduct(productNumber).getId());
        stockOrder.setState(OrderState.DELIVERED);
        stockOrder.setTyp(1);
        stockOrder.setDeliverDate(deliverDate);
        stockOrder.setOrderDate(orderDate);
        INVENTORY_SYSTEM.addStockOrder(stockOrder);

        LOGGER.trace("Successfully ordered '" + amount + "' items " + "of the Product '" + stockProductNumber
                + "' from the Stock.");
        return deliverDate;
    }

    /**
     * Convert ProductNumber from int to String (fixed length: 6 char).
     * @param productNumber productNumber as int
     * @return productNumber as String
     */
    @Override
    public String generateStockProductNumber(final int productNumber) {
        return String.format("%06d", productNumber);
    }

    @Override
    public List<Product> getActiveProducts() {

        final List<Product> products = INVENTORY_SYSTEM.getProductsActive();
        // Update available Amounts (remove reserved items)
        for (final Product product : products) {
            product.setAmount(product.getAmount() - getAmountReserved(product.getNumber()));
        }

        return products;
    }

    /**
     * Returns the amount of reserved items of this product.
     * @param productNumber requested Product
     * @return amount of reserved items
     */
    private int getAmountReserved(final int productNumber) {
        int amountReserved = 0;
        synchronized (RESERVATIONS) {
            for (final ProductReservation reservation : RESERVATIONS) {
                if (reservation.getProductNumber() == productNumber) {
                    amountReserved += reservation.getAmount();
                }
            }
        }

        LOGGER.trace("Currently are " + amountReserved + " items reserved from the productNumber '" + productNumber
                + "'");
        return amountReserved;
    }

    @Override
    public Product getProduct(final int productNumber) {
        final Product product = INVENTORY_SYSTEM.getProduct(productNumber);
        if (product == null) {
            return null;
        }

        // get reserved item count & update product
        final int amountReserved = getAmountReserved(product.getNumber());
        product.setAmount(product.getAmount() - amountReserved);

        LOGGER.trace("getProduct #" + productNumber + ", " + "currently are " + product.getAmount()
                + " items available (" + amountReserved + " are reserved, " + "Total: "
                + (product.getAmount() + amountReserved) + ").");
        return product;
    }

    /**
     * Returns a ProductReservation.
     * @param reservationId requested Reservation ID
     * @return ProductReservation
     */
    private ProductReservation getProductReservation(final int reservationId) {

        for (final ProductReservation reservation : RESERVATIONS) {
            if (reservation.getId() == reservationId) {
                return reservation;
            }
        }

        return null;
    }

    @Override
    public List<Product> getProducts() {

        final List<Product> products = INVENTORY_SYSTEM.getProducts();
        // Update available Amounts (remove reserved items)
        for (final Product product : products) {
            product.setAmount(product.getAmount() - getAmountReserved(product.getNumber()));
        }

        return products;
    }

    @Override
    public boolean productActive(final int productNumber) {
        return INVENTORY_SYSTEM.productActive(productNumber);
    }

    @Override
    public boolean productExists(final int productNumber) {
        return INVENTORY_SYSTEM.productExists(productNumber);
    }



    @Override
    public int reservateProduct(final int productNumber, final int amount) throws ReservationNotPossible {
        // verify Product
        if (productNumber <= 0 || amount <= 0) {
            throw new ReservationNotPossible(productNumber);
        }

        if (!(productExists(productNumber) && productActive(productNumber))) {
            throw new ReservationNotPossible(productNumber);
        }

        final ProductReservation reservation = new ProductReservation(productNumber, amount);
        synchronized (RESERVATIONS) {
            RESERVATIONS.add(reservation);
        }

        LOGGER.trace("New reservation for ProductNumber '" + productNumber + "': " + amount + " items." + "(ID: "
                + reservation.getId() + ")");
        return reservation.getId();
    }

    @Override
    public boolean updateProduct(final Product product) {
        return INVENTORY_SYSTEM.updateProduct(product);
    }

    @Override
    public boolean updateProductReservation(final int reservationId, final int amount)
            throws ReservationNotExistsException {

        synchronized (RESERVATIONS) {
            final ProductReservation reservation = getProductReservation(reservationId);

            if (reservation == null) {
                throw new ReservationNotExistsException(reservationId);
            }
            reservation.setAmount(amount);
        }

        LOGGER.trace("Update reservationId: " + reservationId + ", " + amount + " items.");
        return true;
    }

    /**
     * Updates the Stock, StockCount.
     * @param productNumber the product-number
     * @param newStock new Stock info
     * @return true if success otherwise false
     */
    @Override
    public boolean updateStock(final int productNumber, final int newStock) {
        return INVENTORY_SYSTEM.updateStock(productNumber, newStock);
    }

}
