package ch.hslu.appe.order.logic;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import ch.hslu.appe.inventory.logic.InventorySystem;
import ch.hslu.appe.inventory.logic.InventorySystemImpl;
import ch.hslu.appe.inventory.logic.exception.ReservationNotExistsException;
import ch.hslu.appe.inventory.logic.exception.ReservationNotPossible;
import ch.hslu.appe.inventory.model.Product;
import ch.hslu.appe.order.logic.exception.OrderCreationFailedException;
import ch.hslu.appe.order.logic.exception.OrderException;
import ch.hslu.appe.order.logic.exception.OrderLockedException;
import ch.hslu.appe.order.logic.exception.OrderNotCheckedOutException;
import ch.hslu.appe.order.logic.exception.OrderNotExistsException;
import ch.hslu.appe.order.logic.exception.OrderNotValidException;
import ch.hslu.appe.order.logic.model.UserOrder;
import ch.hslu.appe.order.model.Bill;
import ch.hslu.appe.order.model.Order;
import ch.hslu.appe.order.model.OrderPosition;
import ch.hslu.appe.order.model.OrderState;
import ch.hslu.appe.order.persistency.OrderPersistency;
import ch.hslu.appe.order.persistency.OrderPersistencyImpl;
import ch.hslu.appe.persistency.Persistency;
import ch.hslu.appe.usermanagement.logic.UserManagement;
import ch.hslu.appe.usermanagement.logic.UserManagementImpl;
import ch.hslu.appe.usermanagement.model.User;

/**
 * Order System Implementation.
 */
public final class OrderSystemImpl implements OrderSystem {

    private static final InventorySystem INVENTORY = new InventorySystemImpl();
    private static final Logger LOGGER = LogManager.getLogger("OrderLogic");
    private static final ArrayList<UserOrder> ORDER_CACHE = new ArrayList<UserOrder>();
    private static final OrderPersistency ORDERPERSISTENCY = new OrderPersistencyImpl();
    private static final UserManagement USER_MANAGEMENT = new UserManagementImpl();

    private int nextOrderNumber = -1;
    private final Map<OrderPosition, Integer> productReservations = new HashMap<OrderPosition, Integer>();

    @Override
    public int addPosition(final int userId, final int orderNumber, final int productNumber, final int amount)
            throws OrderException {

        // Check the user and the lock state
        checkUser(orderNumber, userId);

        // Check the product
        final OrderPosition position = new OrderPosition();

        // Get the product from the inventory system
        final Product product = INVENTORY.getProduct(productNumber);
        position.setProduct(product);
        position.setAmount(amount);
        position.setPrice(product.getPrice());
        position.setOutstanding(true);

        synchronized (ORDER_CACHE) {

            final Order order = getCachedOrder(orderNumber);

            // Can only change non persisted positions
            if (order.getId() != 0) {
                throw new OrderException("Order has already been submitted once."
                        + "Its not possible mutate the positions after insertion.");
            }

            // Do a product reservation
            int reservation = 0;
            try {
                reservation = INVENTORY.reservateProduct(productNumber, amount);
            } catch (final ReservationNotPossible e) {
                throw new OrderException("Cant reservate the product!");
            }

            productReservations.put(position, Integer.valueOf(reservation));

            // Set position number
            position.setPosition(getNextPositionNumber(order));

            if (order.getOrderPositions() == null) {
                order.setOrderPositions(new ArrayList<OrderPosition>());
            }

            // Add the position
            order.getOrderPositions().add(position);
        }

        // Add trace message
        LOGGER.trace("Position '" + position.getPosition() + "' added to the order '" + orderNumber + "'");

        return position.getPosition();
    }

    @Override
    public boolean annulateOrder(final int userId, final int number) throws OrderException {

        synchronized (ORDER_CACHE) {

            final UserOrder userOrder = getCachedUserOrder(number);

            if (userOrder == null) {

                // Try to delete the persisted order
                return ORDERPERSISTENCY.deleteOrderByNumber(number);
            } else if (userOrder.getOrder().getId() != 0) {

                // Check the lock
                checkUser(number, userId);

                // Delete from db
                if (ORDERPERSISTENCY.deleteOrderByNumber(number)) {

                    // Delete all positions
                    deleteOrderPositions(userId, userOrder.getOrder());

                    return ORDER_CACHE.remove(userOrder);
                }
            } else {

                // Check the lock
                checkUser(number, userId);

                // Delete all positions
                deleteOrderPositions(userId, userOrder.getOrder());

                // Just remove it from the cache (Not persisted yet)
                return ORDER_CACHE.remove(userOrder);
            }
        }

        LOGGER.error("Failed to annulate the order '" + number + "'!");

        return false;
    }

    @Override
    public int createOrder(final int userId, final Order order) throws OrderCreationFailedException {

        // Cannot create an other order, when the user has already locked an
        // order previously
        if (getActiveOrder(userId) != null) {

            // Log
            LOGGER.error("Cant create a new order when the user has already locked an other order previously. "
                    + "Release the locked order first!");

            // Throw exception
            throw new OrderCreationFailedException(
                    "Cant create a new order when the user has already locked an other order previously. "
                            + "Release the locked order first!");
        }

        // Check for null values
        if (order == null) {

            // Add log entry
            LOGGER.error("Cant create order. Submitted order was null.");

            throw new OrderCreationFailedException("Cant create order. Submitted order was null.");
        }

        // Order number shall be zero
        if (order.getNumber() != 0) {

            // Add log entry
            LOGGER.error("Cant create order. Order number wasn't 0.");

            throw new OrderCreationFailedException("Cant create order. Order number wasn't 0.");
        }

        final User user = USER_MANAGEMENT.getUserById(userId);

        if (user == null) {

            // Add log entry
            LOGGER.error("Cant create order. User does not exists.");

            throw new OrderCreationFailedException("Cant create order. User does not exists.");
        }

        // Set the salesman
        order.setSales(user);

        synchronized (ORDER_CACHE) {

            // Generate an order number
            order.setNumber(generateOrderNumber());

            // Add the order
            ORDER_CACHE.add(new UserOrder(order, user));
        }

        LOGGER.trace("Created a new order and locked it!");

        return order.getNumber();
    }

    @Override
    public boolean deletePosition(final int userId, final int orderNumber, final int positionNumber)
            throws OrderException {

        // Check the user and the lock state
        checkUser(orderNumber, userId);

        // Check the existence
        synchronized (ORDER_CACHE) {

            final Order order = getCachedOrder(orderNumber);

            if (order.getId() != 0) {

                // Trace it
                LOGGER.trace("Cant delete a position of an already submitted order!");

                return false;
            }

            final List<OrderPosition> positions = order.getOrderPositions();
            for (int index = 0; index <= positions.size(); index++) {
                final OrderPosition position = positions.get(index);

                if (position.getPosition() == positionNumber) {

                    // Delete product reservation
                    deleteProductReservation(positions.get(index));

                    // Remove it
                    positions.remove(index);

                    LOGGER.trace("Position '" + positionNumber + "' of the order '" + orderNumber
                            + "' has been deleted.");

                    return true;
                }
            }
        }

        LOGGER.trace("Tried to delete a non existent position '" + positionNumber + "' of the order '" + orderNumber
                + "'");

        return false;
    }

    @Override
    public Order getActiveOrder(final int userId) {

        for (final UserOrder userOrder : ORDER_CACHE) {
            if (userOrder.getUser().getId() == userId) {

                LOGGER.trace("Got the active order '" + userOrder.getOrder().getNumber() + "' of the user '"
                        + userOrder.getUser() + "'");

                return userOrder.getOrder();
            }
        }

        LOGGER.trace("No active order found for the user with the id '" + userId + "'");

        return null;
    }

    @Override
    public Order getOrder(final int userId, final int number) throws OrderException {

        // Check the lock
        checkLock(userId, number);

        if (getCachedOrder(number) != null) {
            return getCachedOrder(number);
        }

        final User user = USER_MANAGEMENT.getUserById(userId);

        if (user == null) {

            // Add log entry
            LOGGER.error("Cant create order. User does not exists.");

            return null;
        }

        // Check existence
        final Order order = ORDERPERSISTENCY.getOrder(number);

        if (order == null) {
            throw new OrderNotExistsException(number);
        }

        // Load the order into the cache
        synchronized (ORDER_CACHE) {
            ORDER_CACHE.add(new UserOrder(order, user));
        }

        return order;
    }

    @Override
    public List<Order> getOrders() {

        return ORDERPERSISTENCY.getOrders();
    }

    @Override
    public OrderState getStatus(final int orderNumber) throws OrderNotExistsException {

        synchronized (ORDER_CACHE) {

            final Order order = getCachedOrder(orderNumber);

            if (order != null) {
                return order.getOrderState();
            }
        }

        // Check existence
        final Order order = ORDERPERSISTENCY.getOrder(orderNumber);
        if (order == null) {
            throw new OrderNotExistsException(orderNumber);
        }

        return order.getOrderState();
    }

    @Override
    public void releaseOrder(final int userId, final int orderNumber) throws OrderException {

        synchronized (ORDER_CACHE) {

            // Check the user
            checkUser(orderNumber, userId);

            final UserOrder userOrder = getCachedUserOrder(orderNumber);

            // Remove positions
            if (userOrder.getOrder().getOrderPositions() != null) {
                for (final OrderPosition position : userOrder.getOrder().getOrderPositions()) {
                    deleteProductReservation(position);
                }
            }

            ORDER_CACHE.remove(userOrder);

            LOGGER.trace("Order '" + orderNumber + "' has been released");
        }
    }

    @Override
    public boolean sendBill(final int userId, final int orderNumber, final Bill bill) {
        // TODO
        return false;
    }

    @Override
    public boolean setStatus(final int userId, final int orderNumber, final OrderState state) throws OrderException {

        // Check the user
        checkLock(userId, orderNumber);

        synchronized (ORDER_CACHE) {

            Order order = getCachedOrder(orderNumber);

            if (order != null) {

                order.setOrderState(state);

                LOGGER.trace("Set the state of the order '" + orderNumber + "' to " + state);

                return true;
            }

            // Check existence
            order = ORDERPERSISTENCY.getOrder(orderNumber);
            if (order == null) {
                throw new OrderNotExistsException(orderNumber);
            }

            order.setOrderState(state);

            ORDERPERSISTENCY.updateOrder(order);

            return true;
        }
    }

    @Override
    public boolean submitOrder(final int userId, final int number) throws OrderException {

        synchronized (ORDER_CACHE) {

            // Check the user
            checkUser(number, userId);

            final Order order = getCachedOrder(number);

            // Check the validity
            validateOrder(order);

            // Create or update order
            if (order.getId() == 0) {

                Date latestDeliveryDate = null;

                ((Persistency) ORDERPERSISTENCY).startGlobalTransaction();

                try {
                    // Commit product reservations
                    for (final OrderPosition position : order.getOrderPositions()) {

                        final Date positionDeliveryDate = INVENTORY
                                .commitProductReservation(getProductReservation(position));

                        // Remove the reservation after commitment
                        productReservations.remove(position);

                        if (positionDeliveryDate == null) {
                            throw new OrderException("Position '" + position.getPosition() + "' is not orderable!");
                        }

                        if (latestDeliveryDate == null || positionDeliveryDate.after(latestDeliveryDate)) {
                            latestDeliveryDate = positionDeliveryDate;
                        }
                    }

                    // Set delivery date
                    order.setDeliverDate(latestDeliveryDate);

                    // Set the order date
                    order.setOrderDate(new Date());

                    final boolean result = ORDERPERSISTENCY.createOrder(order);

                    // Commit the transaction
                    ((Persistency) ORDERPERSISTENCY).commitGlobalTransaction();

                    return result;
                } catch (final Exception ex) {

                    // Roll back the transaction
                    ((Persistency) ORDERPERSISTENCY).rollbackGlobalTransaction();

                    // Log it
                    LOGGER.error("Cant submit the order '" + order.getNumber() + "'", ex);

                    // Re-throw the exception without losing the stacktrace
                    throw ex;
                }
            } else {
                return ORDERPERSISTENCY.updateOrder(order);
            }
        }
    }

    @Override
    public boolean updateOrder(final int userId, final Order order) throws OrderException {

        if (order == null) {
            LOGGER.warn("Cant update a null-order!");

            return false;
        }

        checkUser(order.getNumber(), userId);

        final Order cachedOrder = getCachedOrder(order.getNumber());

        // Update properties
        cachedOrder.setCustomer(order.getCustomer());
        cachedOrder.setDeliverAddress(order.getDeliverAddress());
        cachedOrder.setDeliverCost(order.getDeliverCost());

        return true;
    }

    @Override
    public boolean updatePosition(final int userId, final int orderNumber, final int positionNumber, final int amount)
            throws OrderException {

        // Check the user and the lock
        checkUser(orderNumber, userId);

        synchronized (ORDER_CACHE) {

            final Order order = getOrder(userId, orderNumber);

            // Can only change positions on non submitted orders
            if (order.getId() != 0) {

                // Trace it
                LOGGER.trace("Cant delete a position of an already submitted order!");

                return false;
            }

            // Lookup for the position
            for (final OrderPosition position : order.getOrderPositions()) {

                if (position.getPosition() == positionNumber) {

                    final int reservationId = getProductReservation(position);

                    try {
                        if (!INVENTORY.updateProductReservation(reservationId, amount)) {
                            return false;
                        }
                    } catch (final ReservationNotExistsException e) {
                        LOGGER.error(e.getMessage(), e);

                        // Convert the ReservationNotExistsException into a orderexception
                        final OrderException exception = new OrderException(e.getMessage());
                        exception.initCause(e);

                        throw exception;
                    }

                    // Update the amount
                    position.setAmount(amount);

                    return true;
                }
            }
        }

        return false;
    }

    /**
     * Checks OrderLocked Status.
     * @param userId UserId
     * @param orderNumber OrderNumber
     * @throws OrderLockedException OrderLockedException
     * @throws OrderNotExistsException OrderNotExistsException
     */
    private void checkLock(final int userId, final int orderNumber) throws OrderLockedException,
            OrderNotExistsException {

        UserOrder userOrder = null;

        synchronized (ORDER_CACHE) {
            userOrder = getCachedUserOrder(orderNumber);
        }

        if (userOrder != null && userOrder.getUser().getId() != userId) {
            throw new OrderLockedException(orderNumber, userOrder.getUser().toString());
        }
    }

    /**
     * Check User Cached State.
     * @param orderNumber OrderNumber
     * @param userId UserId
     * @throws OrderNotCheckedOutException OrderNotCheckedOutException
     * @throws OrderLockedException OrderLockedException
     */
    private void checkUser(final int orderNumber, final int userId) throws OrderNotCheckedOutException,
            OrderLockedException {

        UserOrder userOrder = null;

        synchronized (ORDER_CACHE) {
            userOrder = getCachedUserOrder(orderNumber);
        }

        if (userOrder == null) {
            throw new OrderNotCheckedOutException(orderNumber);
        } else if (userOrder.getUser().getId() != userId) {
            throw new OrderLockedException(orderNumber, userOrder.getUser().toString());
        }
    }

    /**
     * Delete a OrderPosition.
     * @param userId logged on User
     * @param order Order
     * @throws OrderException OrderException
     */
    private void deleteOrderPositions(final int userId, final Order order) throws OrderException {

        // Delete all positions
        if (order.getOrderPositions() != null) {

            final List<OrderPosition> positions = order.getOrderPositions();
            for (int index = positions.size() - 1; index >= 0; index--) {
                deletePosition(userId, order.getNumber(), positions.get(index).getPosition());
            }
        }
    }

    /**
     * Delete a Product Reservation.
     * @param position to be deleted
     * @throws OrderException OrderException
     */
    private void deleteProductReservation(final OrderPosition position) throws OrderException {

        if (productReservations.containsKey(position)) {

            final int reservationId = productReservations.get(position).intValue();

            try {
                INVENTORY.deleteProductReservation(reservationId);
            } catch (final ReservationNotExistsException e) {
                LOGGER.error(e.getMessage(), e);

                // Convert the ReservationNotExistsException into a OrderException
                final OrderException exception = new OrderException(e.getMessage());
                exception.initCause(e);

                throw exception;
            }
        }
    }

    /**
     * Genearte a OrderNumber.
     * @return new OrderNumber
     */
    private int generateOrderNumber() {

        // Initialize at the first time
        if (nextOrderNumber == -1) {
            nextOrderNumber = ORDERPERSISTENCY.getNextOrderNumber();
        } else {
            nextOrderNumber++;
        }

        return nextOrderNumber;
    }

    /**
     * Returns a CachedUserOrder.
     * @param number CachedUserOrder Number
     * @return CachedUserOrder
     */
    private Order getCachedOrder(final int number) {

        // Lookup for the order
        final UserOrder userOrder = getCachedUserOrder(number);

        if (userOrder == null) {
            return null;
        } else {
            return userOrder.getOrder();
        }
    }

    /**
     * Returns a CachedUserOrder.
     * @param number CacheduserOrder Number
     * @return CachedUserOrder
     */
    private UserOrder getCachedUserOrder(final int number) {

        // Lookup for the order
        for (final UserOrder order : ORDER_CACHE) {
            if (order.getOrder() != null && order.getOrder().getNumber() == number) {
                return order;
            }
        }

        return null;
    }

    /**
     * Returns the next Positon Number.
     * @param order Order
     * @return next positionNumber
     */
    private int getNextPositionNumber(final Order order) {

        int number = 1;

        if (order.getOrderPositions() == null) {
            return number;
        }

        for (final OrderPosition item : order.getOrderPositions()) {
            if (number <= item.getPosition()) {
                number = item.getPosition() + 1;
            }
        }

        return number;
    }

    /**
     * Returns a Product Reservation ID.
     * @param position OrderPosition
     * @return productReservation ID
     */
    private int getProductReservation(final OrderPosition position) {

        if (productReservations.containsKey(position)) {
            return productReservations.get(position);
        }

        return -1;
    }

    /**
     * Validate a Order.
     * @param order to be validated.
     * @throws OrderNotValidException OrderNotValidException
     */
    private void validateOrder(final Order order) throws OrderNotValidException {

        final List<String> errors = new ArrayList<String>();

        if (order.getDeliverAddress() == null) {
            errors.add("No delivery address declared");
        }

        if (order.getCustomer() == null) {
            errors.add("No customer declared");
        }

        if (order.getOrderPositions().size() == 0) {
            errors.add("No positions added");
        }

        if (order.getSales() == null) {
            errors.add("No salesman declared");
        }

        if (errors.size() > 0) {

            String joinedErrors = "";
            for (final String error : errors) {

                if (!joinedErrors.isEmpty()) {
                    joinedErrors += "\n";
                }

                joinedErrors += "- " + error;
            }

            throw new OrderNotValidException(order.getNumber(), joinedErrors);
        }
    }

}
