package ch.hslu.appe.order.webservice;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.annotation.Resource;
import javax.jws.HandlerChain;
import javax.jws.WebService;
import javax.xml.soap.SOAPException;
import javax.xml.ws.WebServiceContext;

import ch.hslu.appe.inventory.logic.InventorySystem;
import ch.hslu.appe.inventory.logic.InventorySystemImpl;
import ch.hslu.appe.order.logic.OrderSystem;
import ch.hslu.appe.order.logic.OrderSystemImpl;
import ch.hslu.appe.order.logic.exception.OrderException;
import ch.hslu.appe.order.logic.exception.OrderNotExistsException;
import ch.hslu.appe.order.model.Order;
import ch.hslu.appe.order.model.OrderPosition;
import ch.hslu.appe.order.model.OrderState;
import ch.hslu.appe.order.webservice.dto.OrderDto;
import ch.hslu.appe.usermanagement.exception.IllegalNameException;

/**
 * Order Service Implementation.
 */
@WebService(endpointInterface = "ch.hslu.appe.order.webservice.OrderService")
@HandlerChain(file = "./handler/handlers.xml")
public final class OrderServiceImpl implements OrderService {

    private static final InventorySystem inventorySystem = new InventorySystemImpl();
    private static final OrderSystem orderSystem = new OrderSystemImpl();

    @Resource
    WebServiceContext context;

    @Override
    public int addPosition(final int orderNumber, final int productNumber, final int amount) throws OrderException,
    SOAPException {

        return orderSystem.addPosition(getUserId(), orderNumber, productNumber, amount);
    }

    @Override
    public boolean annulateOrder(final int number) throws OrderException, SOAPException {

        return orderSystem.annulateOrder(getUserId(), number);
    }

    @Override
    public int createOrder(final OrderDto order) throws OrderException, IllegalNameException, SOAPException {

        final Order newOrder = order.toOrder();

        // Update products
        if (newOrder.getOrderPositions() != null) {

            final List<OrderPosition> positions = newOrder.getOrderPositions();
            for (int index = 0; index < positions.size(); index++) {
                positions.get(index).setProduct(
                        inventorySystem.getProduct(order.getOrderPositions().get(index).getProductNumber()));
            }
        }

        return orderSystem.createOrder(getUserId(), newOrder);
    }

    @Override
    public boolean deletePosition(final int orderNumber, final int positionNumber)
            throws OrderException, SOAPException {

        return orderSystem.deletePosition(getUserId(), orderNumber, positionNumber);
    }

    @Override
    public OrderDto getActiveOrder() throws SOAPException {

        return OrderDto.fromOrder(orderSystem.getActiveOrder(getUserId()));
    }

    @Override
    public OrderDto getOrder(final int number) throws OrderException, SOAPException {

        return OrderDto.fromOrder(orderSystem.getOrder(getUserId(), number));
    }

    @Override
    public List<OrderDto> getOrders() {

        final List<OrderDto> orders = new ArrayList<OrderDto>();

        for (final Order order : orderSystem.getOrders()) {
            orders.add(OrderDto.fromOrder(order));
        }

        return orders;
    }

    @Override
    public OrderState getStatus(final int orderNumber) throws OrderNotExistsException {

        return orderSystem.getStatus(orderNumber);
    }

    /**
     * Returns the UserID.
     * @return UserId
     * @throws SOAPException SOAPException
     */
    private int getUserId() throws SOAPException {
        final Map<String, Object> items = context.getMessageContext();

        if (items.containsKey("userid")) {
            return (int) items.get("userid");
        }

        throw new SOAPException("Unable to get the user id.");
    }

    @Override
    public void releaseOrder(final int orderNumber) throws OrderException, SOAPException {

        orderSystem.releaseOrder(getUserId(), orderNumber);
    }

    @Override
    public boolean setStatus(final int orderNumber, final OrderState state) throws OrderException, SOAPException {

        return orderSystem.setStatus(getUserId(), orderNumber, state);
    }

    @Override
    public boolean submitOrder(final int number) throws OrderException, SOAPException {

        return orderSystem.submitOrder(getUserId(), number);
    }

    @Override
    public boolean updateOrder(final OrderDto order) throws OrderException, IllegalNameException, SOAPException {

        return orderSystem.updateOrder(getUserId(), order.toOrder());
    }

    @Override
    public boolean updatePosition(final int orderNumber, final int positionNumber, final int amount)
            throws OrderException, SOAPException {

        return orderSystem.updatePosition(getUserId(), orderNumber, positionNumber, amount);
    }
}
