package ch.hslu.appe.usermanagement.logic.session;

import java.io.UnsupportedEncodingException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import ch.hslu.appe.usermanagement.model.User;

/**
 * Session Manager.
 */
public final class SessionManager {

    private static final int DEFAULT_EXPIRE_TIME = 30;
    private static SessionManager instance;
    static final Logger LOGGER = LogManager.getLogger("AccountingLogic");

    /**
     * Gets the singleton instance.
     * @return The instance
     */
    public static synchronized SessionManager getInstance() {
        // Create a new session manager if necessary
        if (instance == null) {
            instance = new SessionManager();
        }

        return instance;
    }

    private final Object lockObject = new Object();

    private final ArrayList<Session> sessions = new ArrayList<Session>();

    /**
     * Private constructor of the SessionManager.
     */
    private SessionManager() {

    }

    /**
     * Clears all sessions.
     */
    public void clear() {

        // Clear all sessions
        synchronized (lockObject) {
            sessions.clear();
        }
    }

    /**
     * Returns the amount of saved sessions.
     * @return Returns amount of valid sessions
     */
    public int count() {

        // Clear out dated sessions
        synchronized (lockObject) {
            for (final Session session : sessions) {
                isValid(session);
            }
        }

        return sessions.size();
    }

    /**
     * Creates a new session for a user which lasts for 30 minutes by default.
     * @param user The user
     * @return Returns the session
     */
    public Session createSession(final User user) {

        return createSession(user, DEFAULT_EXPIRE_TIME);
    }

    /**
     * Creates a new session for a user. If the user has already a valid session, it returns this one.
     * @param user The user
     * @param expires The expire time in minutes
     * @return Returns the session
     */
    public Session createSession(final User user, final int expires) {

        // Check for empty users
        if (user == null) {
            LOGGER.warn("Cannot create a session for an empty user (null) !");

            return null;
        }

        // If minutes are wrong, use a default value of 30 minutes
        if (expires <= 0) {

            LOGGER.trace("Timeout is not valid. " + "Set default timeout for the session (" + DEFAULT_EXPIRE_TIME
                    + " minutes)");
        }

        // Lookup for existent sessions
        Session session = getUserSession(user.getLogin());

        // If there's still a valid user session, delete the last one to ensure, that the user can only have one session
        // at a time
        if (session != null) {
            deleteSession(session.getKey());
        }

        final Date expireDate = new Date((new Date()).getTime() + (expires * 60000L));

        // Create a user session
        session = new Session(user, generateUniqueKey(user.getLogin()), expireDate);

        // Append the session
        synchronized (lockObject) {
            sessions.add(session);
        }

        LOGGER.trace("New session '" + session.getKey() + "' has been created for the user " + user);

        return session;
    }

    /**
     * Remove a Session based on the SessionKey.
     * @param key SessionKey
     * @return true if success otherwise false
     */
    public boolean deleteSession(final String key) {

        final Session session = getSession(key);

        // Check the session
        if (session == null) {
            return false;
        }

        // Delete the session
        sessions.remove(session);

        LOGGER.trace("Session '" + key + "' has been deleted!");

        return true;
    }

    /**
     * Returns a Session based on the SessionKey.
     * @param key SessionKey
     * @return Session
     */
    public Session getSession(final String key) {

        // Lookup for the session
        for (final Session session : sessions) {
            if (session.getKey().equals(key)) {
                return session;
            }
        }

        return null;
    }

    /**
     * Gets the session of a user.
     * @param login The user login
     * @return Returns the session if it has been found, otherwise null
     */
    public Session getUserSession(final String login) {
        Session session = null;

        // Lookup for the session
        synchronized (lockObject) {
            for (final Session item : sessions) {
                if (item.getUser().getLogin().equals(login)) {
                    session = item;
                    break;
                }
            }
        }

        // Return the session (Null if it doesn't exists)
        return session;
    }

    /**
     * Proofs the validity of a session key. If the session is invalid, it removes the session from the internal list.
     * @param session The session instance
     * @return Returns true if the session key is valid, otherwise false
     */
    public boolean isValid(final Session session) {

        if (session == null) {
            return false;
        }

        // If the valid-until attribute is bigger than the current date, the session is valid
        final boolean valid = session.getValidUntil().after((new Date()));

        // Remove invalid sessions
        if (!valid) {
            synchronized (lockObject) {
                sessions.remove(session);

                LOGGER.trace("Removed session '" + session.getKey() + "' due expiration");
            }
        }

        return valid;
    }

    /**
     * Proofs the validity of a session key. If the session is invalid, it removes the session from the internal list.
     * @param key Session key
     * @return Returns true if the session key is valid, otherwise false
     */
    public boolean isValid(final String key) {

        Session session = null;

        // Lookup for the session
        synchronized (lockObject) {
            for (final Session item : sessions) {
                if (item.getKey().equals(key)) {
                    session = item;
                    break;
                }
            }
        }

        // Session has not been found
        if (session == null) {

            LOGGER.warn("Session '" + key + "' does not exists!");

            return false;
        }

        return isValid(session);
    }

    /**
     * Generates a unique id.
     * @param login The login name
     * @return Returns a unique key on success, otherwise null
     */
    private String generateUniqueKey(final String login) {

        try {
            final SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS");

            // Generate a unique string for the hash
            final String data = login + formatter.format(new Date());

            // Create a message digest for sha-1
            final MessageDigest md = MessageDigest.getInstance("SHA");

            // Get the bytes from the data string and hash them
            final byte[] hashedBytes = md.digest(data.getBytes("UTF-8"));

            // Convert the byte-array into a string
            final StringBuffer buffer = new StringBuffer();
            for (int i = 0; i < hashedBytes.length; ++i) {
                buffer.append(Integer.toHexString((hashedBytes[i] & 0xFF) | 0x100).substring(1, 3));
            }

            LOGGER.trace("Unique session key ('" + buffer.toString() + "' generated for the login '" + login + "'");

            return buffer.toString();
        } catch (NoSuchAlgorithmException | UnsupportedEncodingException ex) {

            LOGGER.error("Cannot generate a session key (Login: '" + login + "')", ex);

            return null;
        }
    }
}
