package ch.hslu.appe.webservices.handler;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.Set;

import javax.xml.namespace.QName;
import javax.xml.soap.Node;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPConstants;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPFault;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPMessage;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;
import javax.xml.ws.soap.SOAPFaultException;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import ch.hslu.appe.usermanagement.logic.Access;
import ch.hslu.appe.usermanagement.logic.AccessImpl;
import ch.hslu.appe.usermanagement.logic.session.SessionManager;

/**
 * Access Validator Handler.
 */
public final class AccessValidator implements SOAPHandler<SOAPMessageContext> {

    private static final Access ACCESS = new AccessImpl();
    private static final Logger LOGGER = LogManager.getLogger("Webservice");

    /**
     * attached Error Message for Logger.
     * @param errorMessage error Message to log
     * @param cause cause
     */
    private void attacheErrorMessage(final SOAPMessage errorMessage, final String cause) {
        try {

            final SOAPBody soapBody = errorMessage.getSOAPPart().getEnvelope().getBody();
            final SOAPFault soapFault = soapBody.addFault();
            soapFault.setFaultString(cause);

            throw new SOAPFaultException(soapFault);

        } catch (SOAPException ex) {
            LOGGER.error("SOAP Error", ex);
        }
    }

    @Override
    public void close(final MessageContext context) {
        LOGGER.trace("Service connection has been closed by the client!");
    }

    @Override
    public Set<QName> getHeaders() {
        LOGGER.trace("No headers added to the webservice.");

        return null;
    }

    @Override
    public boolean handleFault(final SOAPMessageContext context) {

        LOGGER.trace("Set handle faults of the webservice to true.");

        return true;
    }

    @Override
    public boolean handleMessage(final SOAPMessageContext context) {

        LOGGER.trace("Received a SOAP message. Try to identify the user.");

        final Boolean isRequest = (Boolean) context.get(MessageContext.MESSAGE_OUTBOUND_PROPERTY);

        // Inbound message from client
        if (!isRequest) {

            try {

                final SOAPMessage soapMsg = context.getMessage();
                final SOAPEnvelope soapEnv = soapMsg.getSOAPPart().getEnvelope();
                SOAPHeader soapHeader = soapEnv.getHeader();

                // if no header, add one
                if (soapHeader == null) {
                    soapHeader = soapEnv.addHeader();
                    attacheErrorMessage(soapMsg, "No SOAP header");

                    LOGGER.error("No SOAP header were submitted in the telegram!");
                }

                // Get client session key from SOAP header
                final Iterator<?> it = soapHeader.extractHeaderElements(SOAPConstants.URI_SOAP_ACTOR_NEXT);

                // If no header block for next actor found? throw exception
                if (it == null || !it.hasNext()) {
                    attacheErrorMessage(soapMsg, "No header");

                    LOGGER.error("Coult not identify the user. No session key has been submitted within the header!");
                }

                // If no session key found? throw exception
                final Node sessionNode = (Node) it.next();
                final String sessionKey = (sessionNode == null) ? null : sessionNode.getValue();

                if (!SessionManager.getInstance().isValid(sessionKey)) {
                    attacheErrorMessage(soapMsg, "Invalid session key. Please login again!");

                    LOGGER.error("The submitted sessionkey '" + sessionKey + "' is not valid!");
                }

                // Check access to method
                if (!ACCESS.checkAccess(sessionKey, soapMsg.getSOAPBody().getFirstChild().getLocalName())) {
                    attacheErrorMessage(soapMsg, "Permission not granted");

                    LOGGER.error("The user does not have permission to access the called webmethod!");
                }

                context.put("userid", SessionManager.getInstance().getSession(sessionKey).getUser().getId());
                context.setScope("userid", MessageContext.Scope.APPLICATION);

                // Log the soap message
                final ByteArrayOutputStream stream = new ByteArrayOutputStream();
                soapMsg.writeTo(stream);
                LOGGER.trace(stream);

                stream.close();

            } catch (SOAPException ex) {

                LOGGER.error("SOAP Error", ex);
            } catch (IOException ex) {

                LOGGER.error("SOAP Error", ex);
            }

        }

        return true;
    }
}
