package ch.hslu.appe.order.logic;

import java.util.ArrayList;
import java.util.Date;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import ch.hslu.appe.inventory.logic.InventorySystem;
import ch.hslu.appe.inventory.logic.InventorySystemImpl;
import ch.hslu.appe.inventory.model.Product;
import ch.hslu.appe.order.logic.exception.OrderCreationFailedException;
import ch.hslu.appe.order.logic.exception.OrderException;
import ch.hslu.appe.order.logic.exception.OrderLockedException;
import ch.hslu.appe.order.logic.exception.OrderNotExistsException;
import ch.hslu.appe.order.model.BillState;
import ch.hslu.appe.order.model.DeliverState;
import ch.hslu.appe.order.model.Order;
import ch.hslu.appe.order.model.OrderState;
import ch.hslu.appe.usermanagement.exception.IllegalNameException;
import ch.hslu.appe.usermanagement.logic.UserManagement;
import ch.hslu.appe.usermanagement.logic.UserManagementImpl;
import ch.hslu.appe.usermanagement.model.Address;
import ch.hslu.appe.usermanagement.model.AddressType;
import ch.hslu.appe.usermanagement.model.Place;
import ch.hslu.appe.usermanagement.model.User;
import ch.hslu.appe.usermanagement.model.UserState;

/**
 * Test OrderSystem.
 */
public final class OrderSystemTest {

    private static final int PRODUCTNUMBER = 10002;

    private Address billingAddress;
    private User customer;
    private Address deliveryAddress;
    private int initialStockAmount;
    private final InventorySystem inventorySystem = new InventorySystemImpl();
    private final OrderSystem orderSystem = new OrderSystemImpl();
    private User salesmanOne;
    private User salesmanTwo;
    private final UserManagement userManagement = new UserManagementImpl();

    /**
     * Returns a TestOrder.
     * @return Order
     */
    private Order getTestOrder() {
        final Order testOrder = new Order();

        testOrder.setCustomer(customer);
        testOrder.setDeliverAddress(deliveryAddress);
        testOrder.setDeliverCost(5.5f);
        testOrder.setSales(salesmanOne);
        testOrder.setPayState(BillState.OPEN);
        testOrder.setOrderState(OrderState.OPEN);
        testOrder.setOrderDate(new Date());
        testOrder.setDeliverState(DeliverState.NONE);
        testOrder.setDeliverDate(new Date());

        return testOrder;
    }

    /**
     * Set Init Values.
     */
    private void resetStock() {
        inventorySystem.updateStock(PRODUCTNUMBER, 30);
    }

    /**
     * JUnit SetUp.
     * @throws IllegalNameException Illegal UserName
     */
    @Before
    public void setup() throws IllegalNameException {

        // Set up the prerequisites for the test
        setupUsers();
        setupAddresses();

        initialStockAmount = inventorySystem.getProduct(PRODUCTNUMBER).getAmount();
    }

    /**
     * Generates TestAddresses.
     */
    private void setupAddresses() {

        final Place placeLucerne = new Place();
        placeLucerne.setZipCode(6002);

        // 1. Add a delivery address
        deliveryAddress = new Address();
        deliveryAddress.setPlace(placeLucerne);
        deliveryAddress.setStreet("Bernerstrasse 543");
        deliveryAddress.setType(AddressType.HOME);
        deliveryAddress.setUser(customer);

        Assert.assertTrue(userManagement.addAddress(customer.getId(), deliveryAddress));

        // 2. Add a billing address
        billingAddress = new Address();
        billingAddress.setPlace(placeLucerne);
        billingAddress.setStreet("Bernerstrasse 543");
        billingAddress.setType(AddressType.BUSINESS);
        billingAddress.setUser(customer);

        Assert.assertTrue(userManagement.addAddress(customer.getId(), billingAddress));

        // Assign the addresses to the customer
        final ArrayList<Address> addresses = new ArrayList<Address>();
        addresses.add(deliveryAddress);
        addresses.add(billingAddress);

        customer.setAddresses(addresses);
    }

    /**
     * Generate Test Users.
     * @throws IllegalNameException Illegal UserName
     */
    private void setupUsers() throws IllegalNameException {
        // 1. Insert a customer
        customer = new User();
        customer.setLogin("john.customer");
        customer.setPassword("1234");
        customer.setRole("customer");
        customer.setState(UserState.ACTIVE);
        customer.setPrename("Hans");
        customer.setSurname("Customer");
        customer.setSalutation("Herr");

        Assert.assertTrue(userManagement.addUser(customer));

        // 2. Insert a salesman
        salesmanOne = new User();
        salesmanOne.setLogin("jack.salesman");
        salesmanOne.setPassword("1234");
        salesmanOne.setRole("salesman");
        salesmanOne.setState(UserState.ACTIVE);
        salesmanOne.setPrename("Jack");
        salesmanOne.setSurname("Salesman");
        salesmanOne.setSalutation("Herr");

        Assert.assertTrue(userManagement.addUser(salesmanOne));

        // 3. Insert a second salesman
        salesmanTwo = new User();
        salesmanTwo.setLogin("thomas.salesman");
        salesmanTwo.setPassword("1234");
        salesmanTwo.setRole("salesman");
        salesmanTwo.setState(UserState.ACTIVE);
        salesmanTwo.setPrename("Thomas");
        salesmanTwo.setSurname("Salesman");
        salesmanTwo.setSalutation("Herr");

        Assert.assertTrue(userManagement.addUser(salesmanTwo));
    }

    /**
     * JUnit TearDown.
     */
    @After
    public void tearDown() {

        // Delete the users and its addresses
        Assert.assertTrue(userManagement.deleteUser(customer.getId()));
        Assert.assertTrue(userManagement.deleteUser(salesmanOne.getId()));
        Assert.assertTrue(userManagement.deleteUser(salesmanTwo.getId()));

        inventorySystem.updateStock(PRODUCTNUMBER, initialStockAmount);
    }

    /**
     * Test OrderLocked Exception.
     * @throws OrderException Order-Exception
     */
    @Test(expected = OrderLockedException.class)
    public void testAnnulateLockedOrder() throws OrderException {
        // Add a new order, but never commit it
        final Order order = getTestOrder();

        // Create it, expect to get a proper order number
        Assert.assertTrue(orderSystem.createOrder(salesmanOne.getId(), order) > 0);

        try {
            // The next statement should throw an exception, because the order
            // is locked by another user
            orderSystem.annulateOrder(salesmanTwo.getId(), order.getNumber());
        } finally {
            Assert.assertTrue(orderSystem.annulateOrder(salesmanOne.getId(), order.getNumber()));
        }
    }

    /**
     * Test Annulate non existent Order.
     * @throws OrderException Order Exception
     */
    @Test
    public void testAnnulateNonExistentOrder() throws OrderException {
        Assert.assertFalse(orderSystem.annulateOrder(salesmanOne.getId(), -1));
    }

    /**
     * Test Annulated Order.
     * @throws OrderException Order Exception
     */
    @Test
    public void testAnnulateOrder() throws OrderException {
        // Add a new order, but never commit it
        final Order order = getTestOrder();

        // Create it, expect to get a proper order number
        Assert.assertTrue(orderSystem.createOrder(salesmanOne.getId(), order) > 0);

        Assert.assertTrue(orderSystem.annulateOrder(salesmanOne.getId(), order.getNumber()));
    }

    /**
     * Test Annulate Order and Reset Reservation.
     * @throws OrderException Order Exception
     */
    @Test
    public void testAnnulateOrderResetReservations() throws OrderException {

        final Order order = getTestOrder();

        // Create an order
        orderSystem.createOrder(salesmanOne.getId(), order);

        // Reset the stock amount
        resetStock();

        final int productStockAmount = inventorySystem.getProduct(PRODUCTNUMBER).getAmount();

        // Append a position
        orderSystem.addPosition(salesmanOne.getId(), order.getNumber(), PRODUCTNUMBER, productStockAmount - 1);

        Assert.assertEquals(1, inventorySystem.getProduct(PRODUCTNUMBER).getAmount());

        // Annulate the order
        orderSystem.annulateOrder(salesmanOne.getId(), order.getNumber());

        Assert.assertEquals(productStockAmount, inventorySystem.getProduct(PRODUCTNUMBER).getAmount());
    }

    /**
     * Test Annulate Persited Order.
     * @throws OrderException Order Exception
     */
    @Test(expected = OrderNotExistsException.class)
    public void testAnnulatePersistedOrder() throws OrderException {

        final Order order = getTestOrder();

        // Create an order
        orderSystem.createOrder(salesmanOne.getId(), order);

        // Append a position
        orderSystem.addPosition(salesmanOne.getId(), order.getNumber(), PRODUCTNUMBER, 57);

        Assert.assertTrue(orderSystem.submitOrder(salesmanOne.getId(), order.getNumber()));

        // Annulate the order
        Assert.assertTrue(orderSystem.annulateOrder(salesmanOne.getId(), order.getNumber()));

        // Order should be vanished from the database -> throws an
        // order-does-not-exists exception
        orderSystem.getOrder(salesmanOne.getId(), order.getNumber());
    }

    /**
     * Test Order Creation Failed.
     * @throws OrderException Order Exception
     */
    @Test(expected = OrderCreationFailedException.class)
    public void testCreateNullOrder() throws OrderException {
        orderSystem.createOrder(customer.getId(), null);
    }

    /**
     * Test Create Order.
     * @throws OrderException Order Exception
     */
    @Test
    public void testCreateOrder() throws OrderException {
        // Add a new order, but never commit it
        final Order order = getTestOrder();

        // Create it, expect to get a proper order number
        Assert.assertTrue(orderSystem.createOrder(salesmanOne.getId(), order) > 0);

        orderSystem.releaseOrder(salesmanOne.getId(), order.getNumber());
    }

    /**
     * Test Create two Orders at once.
     * @throws OrderException Order Exception
     */
    @Test(expected = OrderCreationFailedException.class)
    public void testCreateTwoOrdersAtOnce() throws OrderException {

        final Order firstOrder = getTestOrder();
        final Order secondOrder = getTestOrder();

        orderSystem.createOrder(salesmanOne.getId(), firstOrder);

        // The logic must throw an exception at the next statement, because a
        // user can only have one active order at the same time
        try {
            orderSystem.createOrder(salesmanOne.getId(), secondOrder);
        } finally {
            orderSystem.annulateOrder(salesmanOne.getId(), firstOrder.getNumber());
        }
    }

    /**
     * Test Delete Position.
     */
    @Test
    public void testDeletePosition() {
        // TODO
    }

    /**
     * Test DeliverDate calculation.
     * @throws OrderException OrderException
     */
    @Test
    public void testDeliveryDate() throws OrderException {

        final Order order = getTestOrder();

        // Create an order
        orderSystem.createOrder(salesmanOne.getId(), order);

        // Reset the stock amount
        resetStock();

        final Product product = inventorySystem.getProduct(PRODUCTNUMBER);

        // Append a position (Amount is one piece bigger than the stock amount)
        orderSystem.addPosition(salesmanOne.getId(), order.getNumber(), PRODUCTNUMBER, product.getAmount() + 1);

        orderSystem.submitOrder(salesmanOne.getId(), order.getNumber());

        // Check the date (Must be at least one hour later than now)
        final Date now = new Date();
        Assert.assertTrue(now.getTime() + (1000 * 3600) <= order.getDeliverDate().getTime());

        // Annulate the order
        orderSystem.annulateOrder(salesmanOne.getId(), order.getNumber());
    }

    /**
     * Test Get Active Order.
     * @throws OrderException Order Exception
     */
    @Test
    public void testGetActiveOrder() throws OrderException {

        final Order order = getTestOrder();

        orderSystem.createOrder(salesmanOne.getId(), order);

        Assert.assertEquals(orderSystem.getActiveOrder(salesmanOne.getId()), order);

        orderSystem.releaseOrder(salesmanOne.getId(), order.getNumber());
    }

    /**
     * Test Order Creation Failed Exception.
     * @throws OrderException Order Exception
     */
    @Test(expected = OrderCreationFailedException.class)
    public void testInvalidUserCreateOrder() throws OrderException {

        // Add a new order, but never commit it
        final Order order = getTestOrder();

        orderSystem.createOrder(-1, order);
    }

    /**
     * Test Order Number Generation.
     * @throws OrderException Order Exception
     */
    @Test
    public void testOrderNumberGeneration() throws OrderException {

        final Order firstOrder = getTestOrder();
        final Order secondOrder = getTestOrder();

        orderSystem.createOrder(salesmanOne.getId(), firstOrder);
        orderSystem.createOrder(salesmanTwo.getId(), secondOrder);

        // Second number should be increased by one
        Assert.assertEquals(firstOrder.getNumber() + 1, secondOrder.getNumber());

        orderSystem.releaseOrder(salesmanOne.getId(), firstOrder.getNumber());
        orderSystem.releaseOrder(salesmanTwo.getId(), secondOrder.getNumber());
    }

}
